##
## SciNet's SCMP151, 2023
## Introduction to Quantum Computing
## Erik Spence
##


#########################################################


import numpy as np
import pennylane as qml 


#########################################################


def get_angles(x):

    ## Calculate the angles needed to use multi-controlled rotations
    ## for amplitude embedding..
    
    beta1_1 = 2 * np.arcsin(np.sqrt(x[1] ** 2 / (x[0] ** 2 + x[1] ** 2 + 1e-12)))
    beta1_2 = 2 * np.arcsin(np.sqrt(x[3] ** 2 / (x[2] ** 2 + x[3] ** 2 + 1e-12)))
    beta2_1 = 2 * np.arcsin(np.sqrt((x[2] ** 2 + x[3] ** 2) / (x[0] ** 2 + x[1] ** 2 + x[2] ** 2 + x[3] ** 2)))
    
    return beta1_1, beta1_2, beta2_1


#########################################################


dev = qml.device('default.qubit', wires = 2)


#########################################################


@qml.qnode(dev)
def statepreparation(a):

    ## Apply the single qubit gate.
    qml.RY(a[2], wires = 0)

    ## Apply the controlled R_y rotation, for a control of 1.
    qml.CNOT(wires = [0, 1])
    qml.RY(-a[1]/2, wires = 1)
    qml.CNOT(wires = [0, 1])
    qml.RY(a[1]/2, wires = 1)

    ## Apply the controlled R_y rotation, for a control of 0.
    qml.PauliX(wires = 0)
    qml.CNOT(wires = [0, 1])
    qml.RY(-a[0]/2, wires = 1)
    qml.CNOT(wires = [0, 1])
    qml.RY(a[0]/2, wires = 1)
    qml.PauliX(wires = 0)
    
    return qml.state()


#########################################################
